function imya=intadjust(imy,ims)
%
% Compute the best mapping of intensities of image 'imy' relative to image 'ims'.
% Returns the intensity-adjusted image (imya).
% The images should have been aligned before calling this function.

hsize = size(imy,1);
vsize = size(imy,2);
tsize = hsize * vsize;

% Convert to vectors, and number pixels

intens(1,:) = reshape(imy,[1,tsize]);
intens(2,:) = reshape(ims,[1,tsize]);

% Sort according to iy

[intens,index1] = sortrows(intens',1);
intens = intens';
index2(index1(1:tsize)) = 1:tsize;          % Compute index to undo the sorting

% Now loop through all pixels, computing best intensities

ihigh = 1;

while ihigh <= tsize
    
    ilow = ihigh;
    currentiy = intens(1,ihigh);
    
    while ihigh <= tsize  &  intens(1,min(ihigh,tsize)) == currentiy
        ihigh = ihigh + 1;
    end
    ihigh = ihigh - 1;
    
    bestis = mean(intens(2,ilow:ihigh));
    
    while ilow >= 2 & imya(max(ilow-1,1)) > bestis
        ilow = ilow - 1;
        itocorrect = imya(ilow);
        
        while ilow >= 1  &  imya(max(ilow,1)) == itocorrect
            ilow = ilow - 1;
        end
        ilow = ilow + 1;
        bestis = mean(intens(2,ilow:ihigh));
    end
    
    imya(ilow:ihigh) = bestis;   % Intensity-adjusted image, still with pixels ordered by intensity of imy
    
    ihigh = ihigh + 1;
end

imya = imya(index2);                      % Recover the original pixel order...
imya = reshape(imya,[hsize,vsize]);       % ...and put back into rectangular shape
